// virtualmachine.h
// Gregory Rosenblatt

#ifndef VirtualMachine_H_
#define VirtualMachine_H_

#include <iostream>
#include <vector>
#include <assert.h>
#include "utilities.h"

namespace MetaLang
{
    // enumerations describing the bytecode instructions the VM is able to process
    enum opcode
    {
        op_output=0,
        op_expose_state,
        
        // stack
        op_push_const,
        op_push_var,
        op_push_scope,
        op_pop_scope,
        op_assign,
        
        // registers
        op_store,
        op_load,

        // arithmetic
        op_add,
        op_subtract,
        op_multiply,
        op_divide,
        
        // logical
        op_not,
        op_and,
        op_or,
                
        // comparison
        op_equal,
        op_not_equal,
        op_greater,
        op_greater_equal,
        op_less,
        op_less_equal,
        
        // flow control
        op_jump,
        op_jump_if_true,
        op_jump_if_false,
        op_call_func,
        op_end_func,
        
        // system commands
        op_pusharg_const,
        op_pusharg_var,
        op_call_command,

        op_end,

        // not an opcode
        num_codes   // value is the number of opcodes if first opcode has value of 0
    };

    // the basic instruction
    class Instruction
    {
    public:
        Instruction(opcode code) : _code(code)  {}
        Instruction(opcode code, const char* data, size_t dataSize)
            : _code(code), _data(data, dataSize)  {}
        opcode Code() const         { return _code; }
        const char* Data() const    { return _data.Data(); }
    private:
        opcode  _code;
        Buffer  _data;  // additional data
    };

    // the basic script
    class Script
    {
    public:
        Script(const std::vector<Instruction>& instrList, size_t varCount)
            : _instrList(instrList), _varCount(varCount)	{}
        const Instruction* InstrPtr() const { return &_instrList[0]; }
        const Instruction* End() const      { return &(*_instrList.end()); }
        size_t VarCount() const             { return _varCount; }
    private:
        std::vector<Instruction>    _instrList;
        size_t                      _varCount;
    };
    
    // pointers used as non-modifying dynamic references
    typedef const Script*       ScriptRef;
    typedef const Instruction*  InstrRef;

    // traces the call path through a script    
    class CallStack
    {
    public:
        CallStack(InstrRef& current) : _instr(current)  {}
        void Call(InstrRef funcBegin)
        {
            _stack.push_back(_instr+1); // set the return point to the next instruction
            _instr = funcBegin;         // set current instruction to the function starting point
        }
        void Return()
        {
            assert(!_stack.empty());
            _instr = _stack.back();     // return to proper instruction in calling code
            _stack.pop_back();
        }
    private:
        std::vector<InstrRef>   _stack; // stores function return destinations
        InstrRef&               _instr; // reference to current instruction
    };
    
    // our virtual machine
    template <class SystemType>
    class VirtualMachine
    {
    private:
        // example system command taking int arguments, and returning an int
        typedef Callback<SystemType, int, int>  Command;
        typedef std::vector<int>                ArgVector;
    public:
        VirtualMachine(SystemType& system, const typename Command::Func funcList[], size_t numFuncs)
            : _system(system)    { CreateSystemCallbacks(funcList, numFuncs); }
        // a very basic interface
        void Execute(size_t scriptId);
        size_t Load(const Script& script)   { return AddScript(script); }
        // system commands
        int CallCommand(size_t id, const ArgVector& argv);
        // debugging
        void ShowVariableState() const    { _execution.ExposeStackState(std::cout); }
    private:
        // a script's executional state
        class Execution
        {
        public:
            // construct callstack with current instruction
            Execution() : _scriptPtr(0), _instrPtr(0), _instr(0), _callStack(_instr)  {}
            void Execute(VirtualMachine& vm, ScriptRef scriptPtr);
            // initialization
            void SetDataSize(size_t varCount)   { _stack.SetInitialScope(varCount); }
            // data access
            void SetVar(size_t i, char val) { _stack[i] = val; }
            char GetVar(size_t i) const     { return _stack[i]; }
            void ExposeStackState(std::ostream& out) const;
            // system commands
            int CallSystemCommand(VirtualMachine& vm, size_t id)
                { return vm.CallCommand(id, _argv); }
        private:
            ScopeStack<int>     _stack; // initially of type ScopeStack<char>
            ArgVector           _argv;  // filled and passed to system commands
            ScriptRef           _scriptPtr; // executing script
            InstrRef            _instrPtr;  // root instruction
            InstrRef            _instrEnd;  // end of instruction list
            InstrRef            _instr;     // current instruction
            CallStack           _callStack; // traces call path
            int                 _register;  // general purpose register
        };
    private:    // utilities
        size_t AddScript(const Script& script) // add script to list and retrieve id
        { _scriptList.push_back(script); return (_scriptList.size()-1); }
        ScriptRef SelectScript(size_t index) const  // set current script by id
        {
            assert(index < _scriptList.size());  // make sure the id is valid
            return &_scriptList[index];
        }
        void CreateSystemCallbacks(const typename Command::Func funcList[], size_t numFuncs)
        {
            for (size_t i = 0; i != numFuncs; ++i)
                _commandList.push_back(make_callback(funcList[i]));
        }
    private:    // data members
        Execution               _execution;     // since we're single-threaded, we only need one
        std::vector<Script>     _scriptList;
        std::vector<Command>    _commandList;   // system commands
        SystemType&             _system;
    };

    template <class SystemType>    
    void VirtualMachine<SystemType>::Execution::Execute(VirtualMachine& vm, ScriptRef scriptPtr)
    {
        _scriptPtr = scriptPtr;
        SetDataSize(_scriptPtr->VarCount());    // initialize variable data
        _instrPtr = _scriptPtr->InstrPtr();     // initialize root pointer
        _instrEnd = _scriptPtr->End();          // initialize end marker
        _instr = _instrPtr; // set our iterator to the beginning
        while (_instr < _instrEnd)  // ensure pointer stays in-bounds
        {
            switch(_instr->Code())
            {
            // output
            case op_output:
                cout << static_cast<int>(_stack.Top()) << endl;
                _stack.Pop();
                ++_instr;
                break;
            // expose state
            case op_expose_state:
                ExposeStackState(std::cout);
                ++_instr;
                break;
            // stack
            case op_push_const:
                _stack.Push(_instr->Data()[0]);
                ++_instr;
                break;
            case op_push_var:
                //assert(_instr->Data()[0] < _scriptPtr->VarCount());
                _stack.Push(_stack[_instr->Data()[0]]);
                ++_instr;
                break;
            case op_push_scope:
                _stack.PushScope(_instr->Data()[0]);
                ++_instr;
                break;
            case op_pop_scope:
                _stack.PopScope();
                ++_instr;
                break;
            case op_assign:
                //assert(_instr->Data()[0] < _scriptPtr->VarCount());
                _stack[_instr->Data()[0]] = _stack.Top();
                _stack.Pop();
                ++_instr;
                break;
            // registers
            case op_store:
                _register = _stack.Top();
                _stack.Pop();
                ++_instr;
                break;
            case op_load:
                _stack.Push(_register);
                ++_instr;
                break;
            // arithmetic
            case op_add:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() += top;
            }
                ++_instr;
                break;
            case op_subtract:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() -= top;
            }
                ++_instr;
                break;
            case op_multiply:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() *= top;
            }
                ++_instr;
                break;
            case op_divide:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() /= top;
            }
                ++_instr;
                break;
            // logical
            case op_not:
                _stack.Top() = !_stack.Top();
                ++_instr;
                break;
            case op_and:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() && top);
            }
                ++_instr;
                break;
            case op_or:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() || top);
            }
                ++_instr;
                break;        
            // comparison
            case op_equal:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() == top);
            }
                ++_instr;
                break;
            case op_not_equal:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() != top);
            }
                ++_instr;
                break;
            case op_greater:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() > top);
            }
                ++_instr;
                break;
            case op_greater_equal:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() >= top);
            }
                ++_instr;
                break;
            case op_less:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() < top);
            }
                ++_instr;
                break;
            case op_less_equal:
            {
                char top = _stack.Top();
                _stack.Pop();
                _stack.Top() = (_stack.Top() <= top);
            }
                ++_instr;
                break;
            // flow control
            case op_jump:
                _instr = _instrPtr + _instr->Data()[0];
                break;
            case op_jump_if_true:
                if (_stack.Top() != 0)
                    _instr = _instrPtr + _instr->Data()[0];
                else
                    ++_instr;
                _stack.Pop();                
                break;
            case op_jump_if_false:
                if (_stack.Top() == 0)
                    _instr = _instrPtr + _instr->Data()[0];
                else
                    ++_instr;
                _stack.Pop();
                break;
            case op_call_func:
                _callStack.Call(_instrPtr + _instr->Data()[0]);
                break;
            case op_end_func:
                _callStack.Return();
                break;
            // system commands
            case op_pusharg_const:
                _argv.push_back(_instr->Data()[0]);
                ++_instr;
                break;
            case op_pusharg_var:
                _argv.push_back(_stack[_instr->Data()[0]]);
                ++_instr;
                break;
            case op_call_command:
                _register = CallSystemCommand(vm, _instr->Data()[0]);
                _argv.clear();
                ++_instr;
                break;           
            // end
            case op_end:
                _instr = _instrEnd;  // discontinue the loop
                break;
            }
        }
    }

    template <class SystemType>   
    void VirtualMachine<SystemType>::Execution::ExposeStackState(std::ostream& out) const
    {
        size_t numScopes = _stack.ScopeDepth();
        size_t stackSize = _stack.Size();
        size_t index, label, curDepth = 0, curEnd = 0;
        for (index = 0, label = 0; index < stackSize; ++index, ++label) // for the entire stack
        {
            if (index == curEnd)    // for every new scope in the stack
            {
                label = 0;
                out << "Scope Depth: " << curDepth << endl; // output the scope depth
                if (curDepth < numScopes)
                {
                    if (curDepth+1 < numScopes)
                        curEnd = _stack.GetScopeIndex(curDepth+1);  // get end of this scope
                    else
                        curEnd = _stack.Size(); // end of stack
                    ++curDepth; // advance to next scope depth
                }
                else
                    curEnd = 0;
            }
            out << "    " << label << ": " << static_cast<int>(_stack.GetAtDepth(index, 0)) << endl;  // output each value
        }
    }

    template <class SystemType>
    void VirtualMachine<SystemType>::Execute(size_t scriptId)
    {
        _execution.Execute(*this, SelectScript(scriptId)); // select our script by ID
    }
    
    template <class SystemType>
    int VirtualMachine<SystemType>::CallCommand(size_t id, const ArgVector& argv)
    {
        assert(id < _commandList.size());
        return _commandList[id](_system, argv);
    }
}

#endif
